/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */

package MetaTF.Generator.Java;

import MetaTF.Parser.*;
import MetaTF.Panic;
import MetaTF.Types;
import MetaTF.Generator.Java.Util;
import java.util.*;
import java.io.*;

/**********************************************************************************************
 * DataStructureGenerator.
 * Generate the data structures representing each record of each section
 *
 * @see <a href="../../Parser/metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.3 $, $Date: 2001/03/29 15:53:12 $),
 * @author Richard Jones
**/
class DataStructureGenerator implements DTDVisitor {

     /** The FileWriter to which output is to be written **/
    private final FileWriter writer;

    /** The name of the class to be generated **/
    private final String className;

    /** number of distinct field names **/
    private int numFields = 0; 

    /**
     * Allocate a new object to write putRecord methods
     *
     * @param cname the name of the class being generated
     * @param w the FileWrite to which the method definition is to be written
    **/
    DataStructureGenerator(String cname, FileWriter w) {
        className = cname;
        writer = w;
        if(w==null)
            Panic.panic("out: writer must be called before out is called");
    }
 
    /**
     * Visit the symbol table and emit
     * 1. The Trace data structures
     * 2. the high-level read primitives
     * 3. members and constructors for the reader/writer class
     *    The main constructor includes code to build the trace data structures
    **/
    public Object visit(DTD symtab, Object data) {
        numFields = symtab.getFields().size();
        int ns = symtab.getMaxSectionNumber();
        out(
"    public static void init() {                                                \n"+
"        TraceField tf;                                                                   \n"+
"        MetaTFState.init("+(ns+1)+", "+numFields+");                           \n");

        Map sections = symtab.getSections();
        Set keys = new TreeSet(sections.keySet());
        Iterator it = keys.iterator();
        while(it.hasNext()) {
            Integer sno = (Integer)it.next();
            ((Section)sections.get(sno)).accept(this, symtab);
            out("");
        }
        out(
"    }                                                                                    \n");

        return null;        
    }


    /** The number of the section being visited **/
    int sectionNumber = 0;

    /*
     * Visit each section in ascending numerical order of section numbers
     */
    public Object visit(Section section, Object symtab) {
        sectionNumber = section.getNumber();
        int maxTag = section.getMaxTag();
        out(
"      // SECTION "+sectionNumber+"                                                        \n"+
"        MetaTFState.initSection("+sectionNumber+", "+(maxTag+1)+");                       \n"+
"        for(int i=0; i<"+numFields+"; i++)                                                \n"+
"            Section.initLookupList("+sectionNumber+", i);                                 \n");

        Map records = section.getRecords();
        Set keys = new TreeSet(records.keySet());
        Iterator it = keys.iterator();
        while(it.hasNext()) {
            String s = (String)it.next();
            Record r = (Record)records.get(s);
            if(r!=null)
                try {
                    int t = r.getTag();
                    r.accept(this, symtab);
                } catch (NoTagException e) {
                    // OK
                }
        }
        return null;
    }


    /** The field number - used to count through the TraceFields **/
    int fno; 

    /** The record number (i.e.its tag) **/
    int recordTag;
        
    public Object visit(Record record, Object symtab) {
        int numFields = record.getNumNamedFields();
        try {
            recordTag = record.getTag();
            out(
"      // RECORD "+record.getName()+"                                                      \n"+
"        MetaTFState.initRecord("+sectionNumber+", "+recordTag+", "+numFields+");          ");

            fno = 0; 
            record.getField().accept(this, symtab);
            return null;
        } catch (NoTagException e) {
            Panic.panic("DataStructureGenerator.visit finds record without tag, "+ e.getMessage());
            return null;
        }
    }

    public Object visit(NamedField field, Object symtab) {
        String name = field.getName();
        String type = field.getType();
        Value v = field.getProperty("encoding");
        int index = ((DTD)symtab).getFields().indexOf(name);
        if(index<0)
            Panic.panic("CodeGenerator: "+name+" is not in list of fields");

        out(
"        tf = new "+Util.kindify(type)+"TraceField(\""+name+"\", "+ (String)v.accept(this, symtab)+"); \n"+
"        MetaTFState.setField("+sectionNumber+", "+recordTag+", "+fno+", "+index+", tf);               \n");

        fno++;
        return null;
    }
    

    public Object visit(TagField field, Object symtab) {
        int index = ((DTD)symtab).getFields().indexOf("tag");
        if(index<0)
            Panic.panic("CodeGenerator: tag is not in list of fields");

        out(
"        tf = new NumTraceField(\"tag\", MetaTFConstants.None, 0, 0);         \n"+    // Better to have a new TagTraceField() ??
"        MetaTFState.setField("+sectionNumber+", "+recordTag+", "+fno+", "+index+", tf);   \n");

        fno++;
        return null;
    }

    public Object visit(RepeatField field, Object symtab) {
        field.getRepeat().accept(this, symtab);
        field.getRepeatable().accept(this, symtab);
        return null;
    }
    
    public Object visit(TupleField field, Object symtab) {
        Iterator it = field.getFields().iterator();
        while(it.hasNext()) {
            Field f = (Field)it.next();
            f.accept(this, symtab);
        }
        return null;
    }

    public Object visit(BaseOffset baseOffset, Object data) {
        return "MetaTFConstants.BaseOffset, "+baseOffset.getOffset();
    }
    
    public Object visit(Default def, Object data) {
        return "MetaTFConstants.Default, "+def.getDefault();
    }
    
    public Object visit(StringDefault def, Object data) {
        return "MetaTFConstants.Default, \""+def.getDefault()+"\"";
    }
    
    public Object visit(Delta delta, Object data) {
        return "MetaTFConstants.Delta, "+delta.getDelta();
    }
        
    public Object visit(None none, Object data) {
        return "MetaTFConstants.None";
    }
    
    public Object visit(Stride stride, Object data) {
        return "MetaTFConstants.Stride,"+stride.getStride();
    }
    
    public Object visit(SectionStride sectionStride, Object data) {
        String name = Util.classify(sectionStride.getSectionName());
        return "MetaTFConstants.SectionStride, "+name+".getCache(), "+sectionStride.getStride();
    }
    
    public Object visit(SectionOffset sectionOffset, Object data) {
        String name = Util.classify(sectionOffset.getSectionName());
        return "MetaTFConstants.SectionOffset, "+name+".getCache()";
    }

    public Object visit(NumberValue numberValue, Object data) {
        Panic.panic("DataStructureGenerator finds NumberValue");
        return null;
    }

    /**
     * Utility to emit code
     *
     * @param ln one or more lines of code
    **/
    private void out(String ln) {
        try {
            writer.write(ln+"\n"); 
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
    }

} // end of class DataStructureGenerator

