/*
    File: ASTBaseOffset.java
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created
 */

 
package MetaTF.Parser;

import java.util.*;
import MetaTF.Panic;

/**
 * Build the symbol table
 *
 * @see <a href="metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.4 $, $Date: 2000/12/18 16:57:08 $),
 * @author Richard Jones
**/

public class STBuilder implements MetaTFParserVisitor {

  private static final String cvsheader = "$Header: /home/cur/rej/cvs/MetaTF/src/MetaTF/Parser/STBuilder.java,v 1.4 2000/12/18 16:57:08 rej Exp $";

  /** Accumulate error messages **/
  private String errors = "";

  /** Are there any errors? **/
  private boolean hasError = false;

  public Object visit(SimpleNode node, Object data) {
    Panic.panic("STBuilder found a SimpleNode");
    return null;
  }

  /**
   * Visit the abstract syntax tree.
   * Most visit methods return null; the only exceptions are those that return
   * Value objects (these visit subclasses of ASTValue).
   *
   * @param node the root of the abstract syntax tree representing the DTD
   * @param data not used
   * @return null
  **/
  public Object visit(ASTDTD node, Object data) {
    for(int i=0; i < node.jjtGetNumChildren(); i++)
        node.jjtGetChild(i).jjtAccept(this, null);
    if(hasError) {
        System.out.println("Badly formed DTD:");
        System.out.println(errors);
        System.exit(1);
    }
    return null;
  }

  /**
   * @param data null
   * @return null
  **/
  public Object visit(ASTSection0 node, Object data) {
    try {
        DTD symtab = DTD.instance();
        try {
            Section section = symtab.getSection(0);
            for(int i=0; i < node.getNumPropertiesRecords(); i++) 
                node.getPropertyOrRecord(i).jjtAccept(this, section);
            return section;
        } catch (DTDException e) {
            report(e);
        }
    } catch (DTDException e) {
        report(e);
    }
    return null;
  }

  /**
   * @param data not used
   * @return null
  **/
  public Object visit(ASTSection node, Object data) {
    String sname = node.getSectionName();
    int sno = node.getSectionNumber();
    boolean iscached = node.isCached();
    int line = node.getLine();
    try {
        DTD symtab = DTD.instance();
        try {
            Section section = new Section(sname, sno, iscached, line);
            try {
                symtab.addSection(sno, section);
            } catch (DTDException e) {
                report(e);
            }
            for(int i=2; i < node.getNumPropertiesRecords(); i++ )
                node.getPropertyOrRecord(i).jjtAccept(this, section);
            return section;
        } catch (DTDException e) {
            report(e);
        }
    } catch (DTDException e) {
        report(e);
    }
    return null;
  }

  /**
   * @param section the Section in which this record is defined
   * @return null (or record?)
  **/
  public Object visit(ASTRecord node, Object section) {
    String rname = node.getRecordName();
    int line = node.getLine();
    Record record = new Record(rname, line);
    node.getRecordField().jjtAccept(this, record);     
    for(int i=0; i < node.getNumProperties(); i++ )
        node.getProperty(i).jjtAccept(this, record);  
    try {
	    Section s0 = DTD.instance().getSection(0);
	    Section s = (Section)section;
        s.addRecord(record);
        record.replaceFields(s0.getRecords(), s.getRecords(), new HashMap(s.getFields()));
        record.checkFields();

    } catch (DTDException e) {
        report(e);
    }
    return null; //record;
  }

  /**
   * @param parent is a Section or a Record
   * @return null?
  **/
  public Object visit(ASTProperty node, Object parent) {
    String type  = node.getTypeName();
    String field = node.getFieldName();
    String prop  = node.getPropertyName();
    int line     = node.getLine();
    Value  val   = (Value)node.getPropertyValue().jjtAccept(this, parent);
    try {
        ((HasProperties)parent).addProperty(type, field, prop, val, line);
    } catch (DTDException e) {
        report(e);
    }
    return null; //new NamedField(type, field, prop, val, line); 
  }
    
  /**
   * @param parent the parent Field
   * @return ??
  **/ 
  public Object visit(ASTNamedField node, Object parent) {
    String name = node.getName();
    int line = node.getLine();
    Field sfld;
    if(name.equals("tag"))
        sfld = new TagField(line);
    else
        sfld = new NamedField(name, line);
    try {
        ((HasFields)parent).addField(sfld);
    } catch (DTDException e) {
        report(e);
    }
    return null; //sfld;
  }

  /**
   * @param parent the parent Field
   * @return ??
  **/
  public Object visit(ASTRptField node, Object parent) {
    RepeatField rptfld = new RepeatField();
    node.getRepeat().jjtAccept(this,rptfld);
    node.getRepeated().jjtAccept(this,rptfld);
    try {
        ((HasFields)parent).addField(rptfld);
    } catch (DTDException e) {
        report(e);
    }
    return null; //rptfld;
  }

  /**
   * @param parent the parent Record
   * @return null?
  **/
  public Object visit(ASTTupleField node, Object parent) {
    TupleField tfld = new TupleField(((Record)parent).getName());
    for(int i=0; i<node.getArity(); i++) 
        node.getField(i).jjtAccept(this, tfld);
    try {
    	((HasFields)parent).addField(tfld);
    } catch (DTDException e) {
        report(e);
    }
    return null; //tfld;
  }

  /**
   * @param data not used
   * @return new None
  **/
  public Object visit(ASTNone node, Object data) {
    return new None();
  }

  /**
   * @param data not used
   * @return new BaseOffset
  **/
  public Object visit(ASTBaseOffset node, Object data) {
    return new BaseOffset(node.getBaseOffset());
  }

  /**
   * @param data not used
   * @return new Delta
  **/
  public Object visit(ASTDelta node, Object data) {
    return new Delta(node.getDelta());
  }

  /**
   * @param data not used
   * @return new Default
  **/
  public Object visit(ASTDefault node, Object data) {
    return new Default(node.getDefault());
  }

  /**
   * @param data not used
   * @return new StringDefault
  **/
  public Object visit(ASTStringDefault node, Object data) {
    return new StringDefault(node.getDefault());
  }

  /**
   * @param data not used
   * @return new SectionOffset
  **/
  public Object visit(ASTSectionOffset node, Object data) {
    return new SectionOffset(node.getSection());
  }

  /**
   * @param data not used
   * @return new Stride
  **/
  public Object visit(ASTStride node, Object data) {
    return new Stride(node.getInitialValue(), node.getStride());
  }

  /**
   * @param data not used
   * @return new SectionStride
  **/
  public Object visit(ASTSectionStride node, Object data) {
    return new SectionStride(node.getSection(), node.getStride());
  }

  /**
   * @param data not used
   * @return new NumberValue
  **/
  public Object visit(ASTNumberValue node, Object data) {
    return new NumberValue(node.getNumber());
  }


  public Object visit(ASTName node, Object data) {
    node.dump("Panic: ");
    Panic.panic("STBuilder finds an ASTName");
    return null;
  }

  public Object visit(ASTNumber node, Object data) {
    node.dump("Panic: ");
    Panic.panic("STBuilder finds an ASTName");
    return null;
  }

  public Object visit(ASTTraceString node, Object data) {
    node.dump("Panic: ");
    Panic.panic("STBuilder finds an ASTTraceString");
    return null;
  }
  void report(DTDException e) {
    errors = errors + e.getMessage() + "\n";
    hasError = true;
  }

  /** @return whether any errors were encountered while building the symbol table **/
  boolean hasErrors() { return hasError; }

  /** @return the errors encountered wile building the symbol table **/
  String getErrors() { return errors; }
}
